Package: keylist
Title: Lightweight List Extensions that Enforce Unique Keys
Version: 1.0.0
Authors@R: 
    person("Luke", "Jenkins", , "luke-jenkins-dev@outlook.com", role = c("aut", "cre", "cph"), comment = c(ORCID = "0000-0002-7206-7242"))
Description: Provides two lightweight keylist S3 classes 'klist' and 'knlist': extensions of list that enforce unique keys, supporting either mixed named/unnamed elements or fully named elements, ensuring predictable key-value access.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3
Suggests: testthat (>= 3.0.0)
Config/testthat/edition: 3
URL: https://lj-jenkins.github.io/keylist/,
        https://github.com/LJ-Jenkins/keylist
BugReports: https://github.com/LJ-Jenkins/keylist/issues
NeedsCompilation: yes
Packaged: 2026-04-23 10:06:53 UTC; lukej
Author: Luke Jenkins [aut, cre, cph] (ORCID:
    <https://orcid.org/0000-0002-7206-7242>)
Maintainer: Luke Jenkins <luke-jenkins-dev@outlook.com>
Repository: CRAN
Date/Publication: 2026-04-27 14:20:02 UTC
Built: R 4.5.2; x86_64-apple-darwin20; 2026-05-01 17:13:22 UTC; unix
Archs: keylist.so.dSYM

# Generated by roxygen2: do not edit by hand

S3method("[",klist)
S3method("[",knlist)
S3method("[<-",knlist)
S3method("[[<-",knlist)
S3method("names<-",klist)
S3method("names<-",knlist)
S3method(as.array,klist)
S3method(as.array,knlist)
S3method(as.data.frame,klist)
S3method(as.data.frame,knlist)
S3method(as.keylist,default)
S3method(as.klist,default)
S3method(as.klist,list)
S3method(as.knlist,default)
S3method(as.knlist,list)
S3method(as.list,klist)
S3method(as.list,knlist)
S3method(as.matrix,klist)
S3method(as.matrix,knlist)
S3method(as.vector,klist)
S3method(as.vector,knlist)
S3method(c,klist)
S3method(c,knlist)
S3method(print,klist)
S3method(print,knlist)
export(as.keylist)
export(as.klist)
export(as.knlist)
export(is.keylist)
export(is.klist)
export(is.knlist)
export(keylist)
export(keylist.append)
export(klist)
export(knlist)
useDynLib(keylist, .registration = TRUE)

<!-- README.md is generated from README.Rmd. Please edit that file -->

# keylist <img id="logo" src="man/figures/logo.png" align="right" width="17%" height="17%" />

<!-- badges: start -->

<!-- badges: end -->

Lightweight `list` extensions that enforce unique keys, ensuring
predictable key-value access.

## Installation

Install the latest release of keylist from CRAN.

``` r
install.packages("keylist")
```

You can install the development version from GitHub.

``` r
# install.packages("pak")
pak::pak("LJ-Jenkins/keylist")
```

## Overview

keylist provides two lightweight keylist S3 classes `klist` and
`knlist`: extensions of `list` that enforce unique keys. `klist` accepts
both unnamed and named elements, while `knlist` only accepts named
elements.

`knlist` behaves similarly to Python dictionaries/dicts and can serve as
a more idiomatic R, copy-on-modify alternative without reference
semantics.

``` r
library(keylist)

# klist - unnamed or named (if names present, they must be unique).
klist(1, a = 2)
#> <keylist::klist>
#> [[1]]
#> [1] 1
#> 
#> $a
#> [1] 2
klist(1, a = 1, a = 2, b = 1, b = 2)
#> Error in `klist()`:
#> ! Names must be unique.
#> Duplicate names: a, b

# knlist - fully named and unique.
knlist(a = 1, b = 2)
#> <keylist::knlist>
#> $a
#> [1] 1
#> 
#> $b
#> [1] 2
knlist(1, b = 2)
#> Error in `knlist()`:
#> ! All elements must be named.
knlist(a = 1, a = 2, b = 1, b = 2)
#> Error in `knlist()`:
#> ! Names must be unique.
#> Duplicate names: a, b

# keylist as more general function.
class(keylist(1))
#> [1] "klist"
class(keylist(a = 1, .named = TRUE))
#> [1] "knlist"
```

As extensions to `list`, keylists can take any input list can. Inputs
lists are not validated as they are not keylists - keylist validation
operates on its own level so for validation of nested elements they must
also be keylists.

``` r
# input list separate entity.
keylist(1, list(a = 1, a = 1))
#> <keylist::klist>
#> [[1]]
#> [1] 1
#> 
#> [[2]]
#> [[2]]$a
#> [1] 1
#> 
#> [[2]]$a
#> [1] 1

# each keylist validates its own level.
keylist(1, keylist(a = 1, a = 1))
#> Error in `klist()`:
#> ! Names must be unique.
#> Duplicate names: a
```

All nested elements of a list can be recursively turned into a keylist
when using one of the `as.*` variants.

``` r
x <- list(1, list(a = 1, a = 1))
as.keylist(x, .recursive = TRUE)
#> Error in `as.klist.list()`:
#> ! Names must be unique.
#> Duplicate names: a
```

## Assignment and Modification

Although lists in R can accept duplicate names, R preferentially chooses
the existing list’s name structure when it comes to assignment.
Therefore if you can guarantee that the keys are unique at creation,
there is no way of assigning a duplicate key using the normal R methods
of `[`, `[[` and `$`. See this example:

``` r
x <- list(1, a = 2, 3, b = 4)
x[1] <- list(x = 1)
x[1] # remains unnamed.
#> [[1]]
#> [1] 1

x["a"] <- list(x = 1)
x[2] # name remains 'a'.
#> $a
#> [1] 1

x[5] <- list(k = 10) # index 5 doesn't exist...
x[5] # yet still remains unnamed.
#> [[1]]
#> [1] 10
```

keylist makes use of the above by not implementing any special
validation tricks - it simply ensures the keys are unique upon creation
then let’s you go on your way.

However, assigning by position could enable a non-named element for a
`knlist`. To circumvent this (and to give a clearer definition of type),
`knlist` objects only allow assignment when the indexing is done by
name.

``` r
x <- knlist(a = 1)
x[[1]] <- 1
#> Error:
#> ! Only character indexing is allowed for assignment into knlist objects

# extraction by index is allowed.
x[[1]]
#> [1] 1
```

After creation, the main way that duplicate names may creep into lists
is with direct name editing with `names()` or with list concatenation
with `c()`. keylist includes S3 methods for both `klist` and `knlist`
for these functions to ensure that duplicate keys are not added from
these methods.

``` r
x <- klist(1, a = 1)
names(x) <- c("a", "a")
#> Error in `names<-.klist`:
#> ! Names must be unique.
#> Duplicate names: a

x <- knlist(a = 1)
# name removal allowed for klist, but not knlist.
names(x) <- NULL
#> Error in `names<-.knlist`:
#> ! Names cannot be removed from a knlist object.

# knlist doesn't accept "" or NA names (like klist does).
setNames(x, NA)
#> Error in `names<-.knlist`:
#> ! Names cannot be <NA>.

class(c(klist(1, a = 2), 2, list(3)))
#> [1] "klist"
c(klist(1, a = 2), a = 2, list(3))
#> Error in `c.klist()`:
#> ! Names must be unique.
#> Duplicate names: a
```

## Note

All name comparison is done using C’s `strcmp` function.

## Getting help

If you encounter a clear bug, please file an issue with a minimal
reproducible example on
[GitHub](https://github.com/LJ-Jenkins/keylist/issues).

## Code of Conduct

Please note that the keylist project is released with a [Contributor
Code of
Conduct](https://lj-jenkins.github.io/keylist/CODE_OF_CONDUCT.html). By
contributing to this project, you agree to abide by its terms.

#' Create a List With Unique Keys (Named or Unnamed)
#'
#' @description
#' Create a named/unnamed list with unique keys, erroring if any duplicate
#' keys (names) are found.
#' @param ...
#' for `klist`: objects, possibly named, for `as.klist`: passed to other
#' methods.
#' @param x
#' object to be coerced or tested.
#' @param .recursive
#' boolean indicating whether to recursively validate nested lists.
#' If `TRUE` then all nested lists will be validated and converted to
#' klists, if `FALSE` then only the top level list will be validated
#' and converted to a klist.
#' @param value
#' a character vector of up to the same length as x, or NULL. Resulting names
#' must be unique.
#' @details
#' `names<-` method for `klist` objects will apply the names and then validate
#' that the resulting names are unique.
#'
#' `c.klist` method will combine the objects and then validate that the
#' resulting klist's names are unique.
#'
#' For a list with unique keys of all names, see [knlist].
#' @seealso
#' [knlist] and [keylist].
#' @note
#' klists compare names using C's `strcmp` function.
#'
#' [as.list] and [as.vector] methods for klist objects remove the class
#' and return a base R list or vector.
#' @return
#' A list object of class `klist`. For `is.klist()` a boolean.
#' @examples
#' klist(a = 1, 2, b = 3)
#' try(klist(1, a = 2, a = 1)) # duplicate keys not allowed
#'
#' # objects within a klist are not subject to validation
#' klist(1, list(a = 1, a = 2))
#' try(klist(1, klist(a = 1, a = 2))) # but nested klists are
#'
#' # recursively validate and convert to klist
#' x <- list(1, list(1, 2))
#' x <- as.klist(x, .recursive = TRUE)
#' class(x[[2]]) # nested list is now a klist
#'
#' is.klist(klist(1)) # TRUE
#'
#' try(names(x) <- c("a", "a")) # names are validated when changed
#'
#' # c() method for klist objects also validates
#' try(c(klist(a = 1), list(a = 3)))
#' @export
klist <- function(...) {
  .Call(validate_klist_node_c, list(...))
}

#' @rdname klist
#' @export
is.klist <- function(x) inherits(x, "klist")

#' @rdname klist
#' @export
as.klist <- function(x, ...) {
  UseMethod("as.klist")
}

#' @rdname klist
#' @export
as.klist.default <- function(x, ..., .recursive = FALSE) {
  x <- as.list(x, ...)
  if (isTRUE(.recursive)) {
    .Call(validate_klist_list_c, x)
  } else {
    .Call(validate_klist_node_c, x)
  }
}

#' @export
as.klist.list <- function(x, ..., .recursive = FALSE) {
  if (isTRUE(.recursive)) {
    .Call(validate_klist_list_c, x)
  } else {
    .Call(validate_klist_node_c, x)
  }
}

#' @export
as.list.klist <- function(x, ...) {
  class(x) <- setdiff(class(x), "klist")
  x
}

#' @export
as.vector.klist <- function(x, mode = "any") {
  if (mode == "any" || mode == "list") {
    as.list(x)
  } else {
    as.vector(as.list(x), mode = mode)
  }
}

#' @export
as.data.frame.klist <- function(x, ...) {
  as.data.frame(as.list(x), ...)
}

#' @export
as.array.klist <- function(x, ...) {
  as.array(as.list(x), ...)
}

#' @export
as.matrix.klist <- function(x, ...) {
  as.matrix(as.list(x), ...)
}

#' @export
print.klist <- function(x, ...) {
  x <- as.list(x) # to get rid of the class attr
  cat("<keylist::klist>\n")
  NextMethod()
}

# character and numeric indexing so base list methods fine.

# no need to offer <- methods as after creation the
# assigning into list is always preferred whether named or not.
# e.g., x[3] <- list(name = 1) will be unnamed even
# if didn't exist prior.
# Therefore no dupluicates can be created.

# force class to be retained after [ extraction, so that when
# list elements are extracted they are still klist objects.
# [[ and $ extract the value not the container so not required.

#' @export
"[.klist" <- function(x, i, value) {
  .Call(if_list_force_class, NextMethod(), "klist")
}

# this will cause setNames to change too

#' @rdname klist
#' @export
"names<-.klist" <- function(x, value) {
  .Call(validate_klist_node_c, NextMethod())
}

#' @rdname klist
#' @export
c.klist <- function(...) {
  .Call(validate_klist_node_c, NextMethod())
}

test_that("klist creates a klist", {
  x <- klist(a = 1, 2, b = 3)
  expect_s3_class(x, "klist")
  expect_true(is.klist(x))
})

test_that("klist throws an error if any klist elements are duplicates", {
  expect_no_error(klist(1, b = 2))
  expect_no_error(klist(list()))
  expect_error(klist(a = 1, 2, a = 1))
  expect_error(klist(a = klist(1, b = 2, b = 3), b = 2))
})

test_that("klist validation only applies to klist inputs", {
  expect_no_error(klist(a = list(x = 1, x = 2), b = 2))
  expect_error(klist(a = klist(x = 1, x = 2)))
})

test_that("klist accepts single NA or multiple empty names", {
  x <- klist(x = 1, y = 2)
  expect_no_error(setNames(x, c("", "")))
  y <- klist(x = 1)
  expect_no_error(setNames(y, NA))
  expect_error(setNames(y, c(NA, NA)))
})

test_that("as.klist doesn't recurse as default", {
  x <- list(1, list(a = 1, a = 2))
  expect_no_error(as.klist(x))
  expect_false(inherits(x[[2]], "klist"))
})

test_that("as.klist recursively validates and converts to klist", {
  x <- list(1, list(a = 1, a = 2))
  expect_error(as.klist(x, .recursive = TRUE))
  names(x[[2]]) <- NULL
  expect_no_error(x <- as.klist(x, .recursive = TRUE))
  expect_s3_class(x[[2]], "klist")
})

test_that("as.list removes klist class", {
  x <- klist(a = 1, b = 2)
  y <- as.list(x)
  expect_false(inherits(y, "klist"))
  expect_equal(y, list(a = 1, b = 2))
})

test_that("'[' extraction returns klist", {
  x <- klist(a = 1, b = 2, c = 3)
  y <- x[1:2]
  expect_s3_class(y, "klist")
})

test_that("names<- applies klist logic", {
  x <- klist(a = 1, b = 2)
  expect_error(names(x) <- c("a", "a"))
  expect_no_error(names(x) <- NULL)
  expect_no_error(names(x) <- c("", ""))
})

test_that("c() method for klist objects validates", {
  x <- klist(a = 1)
  expect_error(c(x, 1, list(a = 2), 2))
  expect_no_error(c(x, 1, list(b = 2), 2))
  expect_error(c(x, a = 1, list(2), 2))
  expect_no_error(c(x, 1, list(list(a = 2)), 2))
})

test_that("main as.* methods work", {
  x <- klist(a = 1, b = 2)
  y <- list(a = 1, b = 2)
  expect_equal(as.data.frame(x), as.data.frame(y))
  expect_equal(as.numeric(x), as.numeric(y))
  expect_equal(as.character(x), as.character(y))
  expect_equal(as.logical(x), as.logical(y))
  expect_equal(as.matrix(x), as.matrix(y))
  expect_equal(as.vector(x), as.vector(y))
  expect_equal(as.array(x), as.array(y))
})

#' Create a List With Unique Named Keys
#'
#' @description
#' Create a list with unique names/keys, erroring if any duplicate
#' names/keys are found.
#' @param ...
#' for `knlist`: named objects, for `as.knlist`: passed to other
#' methods.
#' @param x
#' object to be coerced or tested.
#' @param .recursive
#' boolean indicating whether to recursively validate nested lists.
#' If `TRUE` then all nested lists will be validated and converted to
#' knlists, if `FALSE` then only the top level list will be validated
#' and converted to a knlist.
#' @param value
#' a character vector the same length as x. Resulting names
#' must be unique.
#' @details
#' `names<-` method for `knlist` objects will apply the names and then validate
#' that the resulting names are unique.
#'
#' `c.knlist` method will combine the objects and then validate that the
#' resulting knlist's names are unique.
#'
#' For a list with unique keys of names and indexes, see [klist].
#' @seealso
#' [klist] and [keylist].
#' @note
#' knlists compare names using C's `strcmp` function.
#'
#' [as.list] and [as.vector] methods for klist objects remove the class
#' and return a base R list or vector.
#' @return
#' A list object of class `knlist`. For `is.knlist()` a boolean.
#' @examples
#' x <- knlist(a = 1, b = 2, c = 3)
#' try(knlist(b = 1, a = 2, a = 1)) # duplicate keys not allowed
#' try(x[[1]] <- 1) # knlist only accepts character indexing for assignment
#'
#' # objects within a knlist are not subject to validation
#' knlist(x = 1, y = list(a = 1, a = 2))
#' try(knlist(x = 1, y = knlist(a = 1, a = 2))) # but nested knlists are
#'
#' # recursively validate and convert to knlist
#' x <- list(a = 1, b = list(x = 1, y = 2))
#' x <- as.knlist(x, .recursive = TRUE)
#' class(x[[2]]) # nested list is now a knlist
#'
#' is.knlist(knlist(a = 1)) # TRUE
#'
#' try(names(x) <- c("a", "a")) # names are validated when changed
#'
#' # c() method for knlist objects also validates
#' try(c(knlist(a = 1), list(a = 3)))
#' @export
knlist <- function(...) {
  .Call(validate_knlist_node_c, list(...))
}

#' @rdname knlist
#' @export
is.knlist <- function(x) inherits(x, "knlist")

#' @rdname knlist
#' @export
as.knlist <- function(x, ...) {
  UseMethod("as.knlist")
}

#' @rdname knlist
#' @export
as.knlist.default <- function(x, ..., .recursive = FALSE) {
  x <- as.list(x, ...)
  if (isTRUE(.recursive)) {
    .Call(validate_knlist_list_c, x)
  } else {
    .Call(validate_knlist_node_c, x)
  }
}

#' @export
as.knlist.list <- function(x, ..., .recursive = FALSE) {
  if (isTRUE(.recursive)) {
    .Call(validate_knlist_list_c, x)
  } else {
    .Call(validate_knlist_node_c, x)
  }
}

#' @export
as.list.knlist <- function(x, ...) {
  class(x) <- setdiff(class(x), "knlist")
  x
}

#' @export
as.vector.knlist <- function(x, mode = "any") {
  if (mode == "any" || mode == "list") {
    as.list(x)
  } else {
    as.vector(as.list(x), mode = mode)
  }
}


#' @export
as.data.frame.knlist <- function(x, ...) {
  as.data.frame(as.list(x), ...)
}

#' @export
as.array.knlist <- function(x, ...) {
  as.array(as.list(x), ...)
}

#' @export
as.matrix.knlist <- function(x, ...) {
  as.matrix(as.list(x), ...)
}

#' @export
print.knlist <- function(x, ...) {
  x <- as.list(x) # to get rid of the class attr
  cat("<keylist::knlist>\n")
  NextMethod()
}

# character indexing for assignment as we've forced named elements.
# Using chr's will also mean a duplicate could never be
# created from assignment.

# allow non-character indexing for non-assignment as will just return
# value, and allows RStudio viewer.

# force class to be retained after [ extraction, so that when
# container elements are extracted they are still knlist objects.
# [[ and $ extract the value not the container so not required.

stop_non_chr_assignment <- function(i) {
  if (!is.character(i)) {
    stop(
      "Only character indexing is allowed for assignment ",
      "into knlist objects",
      call. = FALSE
    )
  }
}

#' @export
"[.knlist" <- function(x, i, value) {
  .Call(if_list_force_class, NextMethod(), "knlist")
}

#' @export
"[<-.knlist" <- function(x, i, value) {
  stop_non_chr_assignment(i)
  NextMethod()
}

#' @export
"[[<-.knlist" <- function(x, i, value) {
  stop_non_chr_assignment(i)
  NextMethod()
}

# this will cause setNames to change too

#' @rdname knlist
#' @export
"names<-.knlist" <- function(x, value) {
  if (is.null(value)) {
    stop("Names cannot be removed from a knlist object.")
  }
  .Call(validate_knlist_node_c, NextMethod())
}

#' @rdname knlist
#' @export
c.knlist <- function(...) {
  .Call(validate_knlist_node_c, NextMethod())
}

test_that("knlist creates a knlist", {
  x <- knlist(a = 1, b = 2)
  expect_s3_class(x, "knlist")
  expect_true(is.knlist(x))
})

test_that("knlist throws an error if any knlist elements are unnamed", {
  expect_error(knlist(1, b = 2))
  expect_error(knlist(list()))
  expect_error(knlist(a = knlist(1), b = 2))
})

test_that("knlist throws an error if any elements are duplicated", {
  expect_error(knlist(a = 1, a = 2))
})

test_that("knlist only applies to knlist inputs", {
  expect_no_error(knlist(a = list(1), b = 2))
  expect_no_error(knlist(a = list(x = 1, x = 2)))
  expect_error(knlist(a = knlist(x = 1, x = 2)))
})

test_that("knlist doesn't accept NA or empty names", {
  x <- knlist(x = 1)
  expect_error(setNames(x, ""))
  x <- knlist(x = 1)
  expect_error(setNames(x, NA))
  x <- knlist(x = 1, y = 2)
  expect_error(names(x) <- c("a", "a"))
})

test_that("knlist only accepts character indexing for assignment", {
  x <- knlist(a = 1, b = 2)
  expect_error(x[[2]] <- 1)
  expect_error(x[2] <- 1)
})

test_that("as.knlist doesn't recurse as default", {
  x <- list(a = 1, b = list(1, 2))
  expect_no_error(as.knlist(x))
  expect_false(inherits(x[[2]], "knlist"))
})


test_that("as.knlist recursively validates and converts to knlist", {
  x <- list(a = 1, list(a = 1, b = 2))
  expect_error(as.knlist(x, .recursive = TRUE))
  x <- setNames(x, c("a", "b"))
  expect_no_error(x <- as.knlist(x, .recursive = TRUE))
  expect_s3_class(x[[2]], "knlist")
})

test_that("as.list removes knlist class", {
  x <- knlist(a = 1, b = 2)
  y <- as.list(x)
  expect_false(inherits(y, "knlist"))
  expect_equal(y, list(a = 1, b = 2))
})

test_that("'[' extraction returns knlist", {
  x <- knlist(a = 1, b = 2, c = 3)
  y <- x[1:2]
  expect_s3_class(y, "knlist")
})

test_that("names<- applies knlist logic", {
  x <- knlist(a = 1, b = 2)
  expect_error(names(x) <- c("a", "a"))
  expect_error(names(x) <- NULL)
  expect_error(names(x) <- c("", ""))
  expect_no_error(names(x) <- c("x", "y"))
})

test_that("c() method for knlist objects validates", {
  x <- knlist(a = 1)
  expect_error(c(x, b = 1, list(a = 2), x = 2))
  expect_error(c(x, a = 1, list(x = 2), y = 2))
  expect_error(c(x, b = 1, list(2), x = 2))
  expect_no_error(c(x, b = 1, list(x = 2), y = 2))
  expect_error(c(x, b = 1, list(list(a = 2)), x = 2))
  expect_no_error(c(x, b = 1, list(z = list(a = 2)), x = 2))
})

test_that("main as.* methods work", {
  x <- knlist(a = 1, b = 2)
  y <- list(a = 1, b = 2)
  expect_equal(as.data.frame(x), as.data.frame(y))
  expect_equal(as.numeric(x), as.numeric(y))
  expect_equal(as.character(x), as.character(y))
  expect_equal(as.logical(x), as.logical(y))
  expect_equal(as.matrix(x), as.matrix(y))
  expect_equal(as.vector(x), as.vector(y))
  expect_equal(as.array(x), as.array(y))
})

#' Create a List With Unique Keys
#'
#' @description
#' Create a list with unique keys (either fully named, [knlist], or a
#' mix of named and unnamed, [klist]), erroring if duplicate keys are found.
#' @param ...
#' for `keylist`: objects, possibly named, for `as.keylist`: passed to other
#' methods.
#' @param .named
#' boolean indicating whether the list should be fully named or not.
#' If `TRUE` then all elements must be named and unique, if `FALSE` then
#' names are not required but if they are present they must be unique.
#' @param x
#' object to be coerced or tested.
#' @param .recursive
#' boolean indicating whether to recursively validate nested lists.
#' If `TRUE` then all nested lists will be validated and converted to
#' keylists, if `FALSE` then only the top level list will be validated
#' and converted to a keylist.
#' @param klst
#' passed to [append]: the keylist to which the values are to be appended.
#' @param values
#' passed to [append]: the values to be included in the modified keylist.
#' @param after
#' passed to [append]: a subscript, after which the values are to be appended.
#' @details
#' `keylist()` creates one of the two keylist objects: [`klist`] or [`knlist`],
#' depending on the value of the `.named` argument. Those methods are
#' generally preferred as they are more explicit.
#'
#' `is.keylist()` checks if an object inherits from either `klist` or `knlist`.
#'
#' `keylist.append()` applies [append] to a keylist and validates the result
#' depending on whether the input was a `klist` or `knlist`.
#'
#' keylist objects accept any R object as elements, but the keys must be
#' unique. The key validation is done at the top level, so nested lists
#' are not validated unless they are also keylists. To recursively turn
#' a nested list structure into keylists, use
#' `as.keylist(x, .recursive = TRUE)` or one of the lower level variants.
#' @seealso
#' [klist] and [knlist].
#' @note
#' keylists compare names using C's `strcmp` function.
#'
#' [as.list] and [as.vector] methods for keylist objects remove the class
#' and return a base R list or vector.
#' @return
#' A list object of class `klist` or `knlist`. For `is.keylist()` a boolean.
#' @examples
#' keylist(a = 1, 2, b = 3) # default is a klist
#' try(keylist(1, a = 2, a = 1)) # duplicate keys not allowed
#'
#' x <- keylist(a = 1, b = 2, .named = TRUE) # create a knlist
#' try(keylist(1, b = 2, .named = TRUE)) # unnamed keys not allowed
#' try(x[[1]] <- 1) # knlist only accepts character indexing for assignment
#'
#' # objects within a keylist are not subject to validation
#' keylist(1, list(a = 1, a = 2))
#' try(keylist(1, keylist(a = 1, a = 2))) # but nested keylists are
#'
#' # recursively validate and convert to keylist
#' x <- list(1, list(1, 2))
#' x <- as.keylist(x, .recursive = TRUE)
#' class(x[[2]]) # nested list is now a keylist
#'
#' is.keylist(klist(1)) && is.keylist(knlist(a = 1)) # TRUE
#'
#' keylist.append(klist(a = 1), list(2, b = 3)) # append to a klist
#' # c() method for keylist objects also validates
#' try(c(keylist(a = 1), list(a = 3)))
#' @export
keylist <- function(..., .named = FALSE) {
  if (isTRUE(.named)) {
    knlist(...)
  } else {
    klist(...)
  }
}

#' @rdname keylist
#' @export
is.keylist <- function(x) inherits(x, c("klist", "knlist"))

#' @rdname keylist
#' @export
as.keylist <- function(x, ...) {
  UseMethod("as.keylist")
}

#' @rdname keylist
#' @export
as.keylist.default <- function(x, ..., .named = FALSE, .recursive = FALSE) {
  if (isTRUE(.named)) {
    as.knlist(x, ..., .recursive = .recursive)
  } else {
    as.klist(x, ..., .recursive = .recursive)
  }
}

#' @rdname keylist
#' @export
keylist.append <- function(klst, values, after = length(klst)) {
  if (!is.keylist(klst)) {
    stop("Only keylist objects can be appended to.")
  }
  cl <- which(c("klist", "knlist") %in% class(klst))
  if (length(cl) > 1) {
    stop("Object cannot be both a klist and a knlist.")
  }

  x <- append(klst, values, after = after)

  switch(cl,
    .Call(validate_klist_node_c, x),
    .Call(validate_knlist_node_c, x)
  )
}

test_that("klist creates a klist/knlist object", {
  x <- keylist(1)
  expect_s3_class(x, "klist")
  y <- keylist(a = 1, .named = TRUE)
  expect_s3_class(y, "knlist")
  expect_true(is.keylist(x))
  expect_true(is.keylist(y))
  expect_error(y[[2]] <- 1)
  expect_error(y[2] <- 1)
})

test_that("as.keylist doesn't recurse as default", {
  x <- list(1, list(a = 1, a = 2))
  expect_no_error(as.keylist(x))
  expect_false(inherits(x[[2]], "klist"))

  x <- list(a = 1, b = list(1, 2))
  expect_no_error(as.keylist(x, .named = TRUE))
  expect_false(inherits(x[[2]], "knlist"))
})

test_that("as.keylist recursively validates and converts to klist", {
  x <- list(1, list(a = 1, a = 2))
  expect_error(as.keylist(x, .recursive = TRUE))
  names(x[[2]]) <- NULL
  expect_no_error(x <- as.keylist(x, .recursive = TRUE))
  expect_s3_class(x[[2]], "klist")

  x <- list(a = 1, list(a = 1, b = 2))
  expect_error(as.keylist(x, .named = TRUE, .recursive = TRUE))
  x <- setNames(x, c("a", "b"))
  expect_no_error(x <- as.keylist(x, .named = TRUE, .recursive = TRUE))
  expect_s3_class(x[[2]], "knlist")
})

test_that("keylist.append appends elements correctly", {
  x <- keylist(1)
  y <- keylist(2)
  z <- keylist.append(x, y)
  expect_s3_class(z, "klist")
  expect_equal(length(z), 2)
})

test_that("keylist.append creates keylist even if values aren't a keylist", {
  z <- keylist.append(klist(1), list(2))
  expect_s3_class(z, "klist")
  expect_equal(length(z), 2)
  z <- keylist.append(klist(1), 1:10)
  expect_s3_class(z, "klist")
  expect_equal(length(z), 11)
})

test_that("keylist.append validates the resulting object", {
  x <- keylist(a = 1)
  expect_error(keylist.append(x, list(a = 2)))
  expect_error(keylist.append(x, c(a = 2)))
  expect_no_error(keylist.append(x, list(1)))
  expect_no_error(keylist.append(x, c(b = 1)))

  x <- keylist(a = 1, .named = TRUE)
  expect_error(keylist.append(x, list(a = 2)))
  expect_error(keylist.append(x, list(1)))
  expect_error(keylist.append(x, 1))
  expect_error(keylist.append(x, c(a = 2)))
  expect_no_error(keylist.append(x, c(b = 2)))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/klist.r
\name{klist}
\alias{klist}
\alias{is.klist}
\alias{as.klist}
\alias{as.klist.default}
\alias{names<-.klist}
\alias{c.klist}
\title{Create a List With Unique Keys (Named or Unnamed)}
\usage{
klist(...)

is.klist(x)

as.klist(x, ...)

\method{as.klist}{default}(x, ..., .recursive = FALSE)

\method{names}{klist}(x) <- value

\method{c}{klist}(...)
}
\arguments{
\item{...}{for \code{klist}: objects, possibly named, for \code{as.klist}: passed to other
methods.}

\item{x}{object to be coerced or tested.}

\item{.recursive}{boolean indicating whether to recursively validate nested lists.
If \code{TRUE} then all nested lists will be validated and converted to
klists, if \code{FALSE} then only the top level list will be validated
and converted to a klist.}

\item{value}{a character vector of up to the same length as x, or NULL. Resulting names
must be unique.}
}
\value{
A list object of class \code{klist}. For \code{is.klist()} a boolean.
}
\description{
Create a named/unnamed list with unique keys, erroring if any duplicate
keys (names) are found.
}
\details{
\verb{names<-} method for \code{klist} objects will apply the names and then validate
that the resulting names are unique.

\code{c.klist} method will combine the objects and then validate that the
resulting klist's names are unique.

For a list with unique keys of all names, see \link{knlist}.
}
\note{
klists compare names using C's \code{strcmp} function.

\link{as.list} and \link{as.vector} methods for klist objects remove the class
and return a base R list or vector.
}
\examples{
klist(a = 1, 2, b = 3)
try(klist(1, a = 2, a = 1)) # duplicate keys not allowed

# objects within a klist are not subject to validation
klist(1, list(a = 1, a = 2))
try(klist(1, klist(a = 1, a = 2))) # but nested klists are

# recursively validate and convert to klist
x <- list(1, list(1, 2))
x <- as.klist(x, .recursive = TRUE)
class(x[[2]]) # nested list is now a klist

is.klist(klist(1)) # TRUE

try(names(x) <- c("a", "a")) # names are validated when changed

# c() method for klist objects also validates
try(c(klist(a = 1), list(a = 3)))
}
\seealso{
\link{knlist} and \link{keylist}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/keylist.r
\name{keylist}
\alias{keylist}
\alias{is.keylist}
\alias{as.keylist}
\alias{as.keylist.default}
\alias{keylist.append}
\title{Create a List With Unique Keys}
\usage{
keylist(..., .named = FALSE)

is.keylist(x)

as.keylist(x, ...)

\method{as.keylist}{default}(x, ..., .named = FALSE, .recursive = FALSE)

keylist.append(klst, values, after = length(klst))
}
\arguments{
\item{...}{for \code{keylist}: objects, possibly named, for \code{as.keylist}: passed to other
methods.}

\item{.named}{boolean indicating whether the list should be fully named or not.
If \code{TRUE} then all elements must be named and unique, if \code{FALSE} then
names are not required but if they are present they must be unique.}

\item{x}{object to be coerced or tested.}

\item{.recursive}{boolean indicating whether to recursively validate nested lists.
If \code{TRUE} then all nested lists will be validated and converted to
keylists, if \code{FALSE} then only the top level list will be validated
and converted to a keylist.}

\item{klst}{passed to \link{append}: the keylist to which the values are to be appended.}

\item{values}{passed to \link{append}: the values to be included in the modified keylist.}

\item{after}{passed to \link{append}: a subscript, after which the values are to be appended.}
}
\value{
A list object of class \code{klist} or \code{knlist}. For \code{is.keylist()} a boolean.
}
\description{
Create a list with unique keys (either fully named, \link{knlist}, or a
mix of named and unnamed, \link{klist}), erroring if duplicate keys are found.
}
\details{
\code{keylist()} creates one of the two keylist objects: \code{\link{klist}} or \code{\link{knlist}},
depending on the value of the \code{.named} argument. Those methods are
generally preferred as they are more explicit.

\code{is.keylist()} checks if an object inherits from either \code{klist} or \code{knlist}.

\code{keylist.append()} applies \link{append} to a keylist and validates the result
depending on whether the input was a \code{klist} or \code{knlist}.

keylist objects accept any R object as elements, but the keys must be
unique. The key validation is done at the top level, so nested lists
are not validated unless they are also keylists. To recursively turn
a nested list structure into keylists, use
\code{as.keylist(x, .recursive = TRUE)} or one of the lower level variants.
}
\note{
keylists compare names using C's \code{strcmp} function.

\link{as.list} and \link{as.vector} methods for keylist objects remove the class
and return a base R list or vector.
}
\examples{
keylist(a = 1, 2, b = 3) # default is a klist
try(keylist(1, a = 2, a = 1)) # duplicate keys not allowed

x <- keylist(a = 1, b = 2, .named = TRUE) # create a knlist
try(keylist(1, b = 2, .named = TRUE)) # unnamed keys not allowed
try(x[[1]] <- 1) # knlist only accepts character indexing for assignment

# objects within a keylist are not subject to validation
keylist(1, list(a = 1, a = 2))
try(keylist(1, keylist(a = 1, a = 2))) # but nested keylists are

# recursively validate and convert to keylist
x <- list(1, list(1, 2))
x <- as.keylist(x, .recursive = TRUE)
class(x[[2]]) # nested list is now a keylist

is.keylist(klist(1)) && is.keylist(knlist(a = 1)) # TRUE

keylist.append(klist(a = 1), list(2, b = 3)) # append to a klist
# c() method for keylist objects also validates
try(c(keylist(a = 1), list(a = 3)))
}
\seealso{
\link{klist} and \link{knlist}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/knlist.r
\name{knlist}
\alias{knlist}
\alias{is.knlist}
\alias{as.knlist}
\alias{as.knlist.default}
\alias{names<-.knlist}
\alias{c.knlist}
\title{Create a List With Unique Named Keys}
\usage{
knlist(...)

is.knlist(x)

as.knlist(x, ...)

\method{as.knlist}{default}(x, ..., .recursive = FALSE)

\method{names}{knlist}(x) <- value

\method{c}{knlist}(...)
}
\arguments{
\item{...}{for \code{knlist}: named objects, for \code{as.knlist}: passed to other
methods.}

\item{x}{object to be coerced or tested.}

\item{.recursive}{boolean indicating whether to recursively validate nested lists.
If \code{TRUE} then all nested lists will be validated and converted to
knlists, if \code{FALSE} then only the top level list will be validated
and converted to a knlist.}

\item{value}{a character vector the same length as x. Resulting names
must be unique.}
}
\value{
A list object of class \code{knlist}. For \code{is.knlist()} a boolean.
}
\description{
Create a list with unique names/keys, erroring if any duplicate
names/keys are found.
}
\details{
\verb{names<-} method for \code{knlist} objects will apply the names and then validate
that the resulting names are unique.

\code{c.knlist} method will combine the objects and then validate that the
resulting knlist's names are unique.

For a list with unique keys of names and indexes, see \link{klist}.
}
\note{
knlists compare names using C's \code{strcmp} function.

\link{as.list} and \link{as.vector} methods for klist objects remove the class
and return a base R list or vector.
}
\examples{
x <- knlist(a = 1, b = 2, c = 3)
try(knlist(b = 1, a = 2, a = 1)) # duplicate keys not allowed
try(x[[1]] <- 1) # knlist only accepts character indexing for assignment

# objects within a knlist are not subject to validation
knlist(x = 1, y = list(a = 1, a = 2))
try(knlist(x = 1, y = knlist(a = 1, a = 2))) # but nested knlists are

# recursively validate and convert to knlist
x <- list(a = 1, b = list(x = 1, y = 2))
x <- as.knlist(x, .recursive = TRUE)
class(x[[2]]) # nested list is now a knlist

is.knlist(knlist(a = 1)) # TRUE

try(names(x) <- c("a", "a")) # names are validated when changed

# c() method for knlist objects also validates
try(c(knlist(a = 1), list(a = 3)))
}
\seealso{
\link{klist} and \link{keylist}.
}
